@@ -2,15 +2,15 @@
  *
  *  Copyright (C) 2000-2011 Thomas Habets <thomas@habets.se>
  *
- *  This library is free software; you can redistribute it and/or
- *  modify it under the terms of the GNU General Public
- *  License as published by the Free Software Foundation; either
- *  version 2 of the License, or (at your option) any later version.
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
  *
- *  This library is distributed in the hope that it will be useful,
+ *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- *  General Public License for more details.
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License along
  *  with this program; if not, write to the Free Software Foundation, Inc.,
@@ -27,7 +27,7 @@
 
 extern int verbose;
 void do_signal_init();
-void do_libnet_init(const char *ifname);
+void do_libnet_init(const char *ifname, int recursive);
 void sigint(int);
 const char *arping_lookupdev_default(uint32_t srcip, uint32_t dstip,
 				     char *ebuf);
